from efootprint.abstract_modeling_classes.explainable_object_base_class import Source, ExplainableObject
from efootprint.abstract_modeling_classes.explainable_objects import EmptyExplainableObject, ExplainableQuantity
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.sources import Sources
from efootprint.constants.units import u
from efootprint.core.hardware.storage import Storage
from efootprint.core.hardware.server_base import ServerBase, ServerTypes

BLOOM_PAPER_SOURCE = Source("Estimating the Carbon Footprint of BLOOM", "https://arxiv.org/abs/2211.05100")


class GPUServer(ServerBase):
    @classmethod
    def default_values(cls):
        return {
            "server_type": ServerTypes.serverless(),
            "gpu_power": SourceValue(400 * u.W / u.gpu, BLOOM_PAPER_SOURCE, "GPU Power"),
            "gpu_idle_power": SourceValue(50 * u.W / u.gpu, BLOOM_PAPER_SOURCE, "GPU idle power"),
            "ram_per_gpu": SourceValue(80 * u.GB / u.gpu, BLOOM_PAPER_SOURCE, label="RAM per GPU"),
            "carbon_footprint_fabrication_per_gpu": SourceValue(
                150 * u.kg / u.gpu, BLOOM_PAPER_SOURCE, "Carbon footprint one GPU"),
            "average_carbon_intensity": SourceValue(100 * u.g / u.kWh),
            "carbon_footprint_fabrication_without_gpu": SourceValue(
            2500 * u.kg, BLOOM_PAPER_SOURCE, "Carbon footprint without GPU"),
            "compute": SourceValue(4 * u.gpu),
            "lifespan": SourceValue(6 * u.year),
            "power_usage_effectiveness": SourceValue(1.2 * u.dimensionless),
            "server_utilization_rate": SourceValue(1 * u.dimensionless),
            "base_compute_consumption": SourceValue(0 * u.gpu),
            "base_ram_consumption": SourceValue(0 * u.GB),
            "fixed_nb_of_instances": EmptyExplainableObject()
            }
    
    def __init__(self, name: str, server_type: ExplainableObject,  gpu_power: ExplainableQuantity,
                 gpu_idle_power: ExplainableQuantity, ram_per_gpu: ExplainableQuantity,
                 carbon_footprint_fabrication_per_gpu: ExplainableQuantity,
                 average_carbon_intensity: ExplainableQuantity, compute: ExplainableQuantity,
                 carbon_footprint_fabrication_without_gpu: ExplainableQuantity, lifespan: ExplainableQuantity,
                 power_usage_effectiveness: ExplainableQuantity, server_utilization_rate: ExplainableQuantity,
                 base_compute_consumption: ExplainableQuantity, base_ram_consumption: ExplainableQuantity, storage: Storage,
                 fixed_nb_of_instances: ExplainableQuantity | EmptyExplainableObject = None):
        super().__init__(
            name, server_type, carbon_footprint_fabrication=SourceValue(0 * u.kg), power=SourceValue(0 * u.W),
            lifespan=lifespan, idle_power=SourceValue(0 * u.W), ram=SourceValue(0 * u.GB),
            compute=compute, power_usage_effectiveness=power_usage_effectiveness,
            average_carbon_intensity=average_carbon_intensity, server_utilization_rate=server_utilization_rate,
            base_compute_consumption=base_compute_consumption, base_ram_consumption=base_ram_consumption,
            storage=storage, fixed_nb_of_instances=fixed_nb_of_instances)
        self.gpu_power = gpu_power.set_label(f"{self.name} GPU power")
        self.gpu_idle_power = gpu_idle_power.set_label(f"{self.name} GPU idle power")
        self.ram_per_gpu = ram_per_gpu.set_label(f"{self.name} RAM per GPU")
        self.carbon_footprint_fabrication_without_gpu = carbon_footprint_fabrication_without_gpu.set_label(
            f"{self.name} carbon footprint without GPU")
        self.carbon_footprint_fabrication_per_gpu = carbon_footprint_fabrication_per_gpu.set_label(
            f"{self.name} carbon footprint one GPU")

    @property
    def calculated_attributes(self):
        return ["carbon_footprint_fabrication", "power", "idle_power", "ram"] + super().calculated_attributes

    def update_carbon_footprint_fabrication(self):
        self.carbon_footprint_fabrication = (self.carbon_footprint_fabrication_without_gpu
                + self.compute * self.carbon_footprint_fabrication_per_gpu
                ).set_label(f"{self.name} carbon footprint fabrication")

    def update_power(self):
        self.power = (self.gpu_power * self.compute).set_label(f"{self.name} power")

    def update_idle_power(self):
        self.idle_power = (self.gpu_idle_power * self.compute).set_label(f"{self.name} idle power")

    def update_ram(self):
        self.ram = (self.ram_per_gpu * self.compute).set_label(f"{self.name} RAM")



from efootprint.logger import logger


def upgrade_version_9_to_10(system_dict):
    if "Hardware" in system_dict.keys():
        logger.info(f"Upgrading system dict from version 9 to 10, changing 'Hardware' key to 'Device'")
        system_dict["Device"] = system_dict.pop("Hardware")

    return system_dict


VERSION_UPGRADE_HANDLERS = {
    9: upgrade_version_9_to_10
}



from efootprint.utils.graph_tools import set_string_max_width


def nodes_at_depth(node, depth=0, depth_lists=None):
    if depth_lists is None:
        depth_lists = {}

    if node.label:
        if depth not in depth_lists:
            depth_lists[depth] = []
        for i in range(0, depth):
            depth_lists[i] = [n for n in depth_lists[i] if n.label != node.label]
        if node.label not in [n.label for n in depth_lists[depth]]:
            depth_lists[depth].append(node)

        depth += 1

    if node.left_parent:
        nodes_at_depth(node.left_parent, depth, depth_lists)
    if node.right_parent:
        nodes_at_depth(node.right_parent, depth, depth_lists)

    return depth_lists


def calculate_positions(node):
    depth_lists = nodes_at_depth(node)
    max_width = max(len(lst) for lst in depth_lists.values())
    max_depth = len(depth_lists.keys())
    pos = {}

    for depth, nodes in depth_lists.items():
        num_nodes = len(nodes)
        for i, n in enumerate(nodes):
            offset = (num_nodes - 1) / 2
            x = (i - offset) * (max_width / num_nodes)
            pos[n.label] = (x, max_depth - depth)

    return pos


def build_calculus_graph(
        root_node, colors_dict=None, x_multiplier=150, y_multiplier=150, width="1800px", height="900px",
        notebook=False, max_depth=100):
    from pyvis.network import Network

    if colors_dict is None:
        colors_dict = {"user data": "gold", "default": "darkred"}
    cdn_resources = "local"
    if notebook:
        cdn_resources = "in_line"

    G = Network(notebook=notebook, directed=True, width=width, height=height, cdn_resources=cdn_resources)
    G.toggle_physics(False)

    pos = calculate_positions(root_node)

    def add_nodes_edges(node, current_depth, parent_id=None):
        if node.label:
            if node.modeling_obj_container:
                current_depth += 1
            color = None
            if node.left_parent is None and node.right_parent is None:
                if getattr(node, "source", None) is not None:
                    source_name = node.source.name
                    if source_name in colors_dict.keys():
                        color = colors_dict[source_name]
                    else:
                        color = colors_dict["default"]
            G.add_node(
                node.label, label=set_string_max_width(node.label, 20),
                title=set_string_max_width(str(node.explain()), 80),
                x=pos[node.label][0]*x_multiplier, y=pos[node.label][1]*y_multiplier, color=color, size=15)
            if parent_id:
                G.add_edge(node.label, parent_id)
            current_id = node.label
        else:
            current_id = parent_id

        if node.left_parent and current_depth <= max_depth:
            add_nodes_edges(node.left_parent, current_depth, current_id)
        if node.right_parent and current_depth <= max_depth:
            add_nodes_edges(node.right_parent, current_depth, current_id)

    add_nodes_edges(root_node, 0)

    return G



from copy import copy

from efootprint.abstract_modeling_classes.contextual_modeling_object_attribute import ContextualModelingObjectAttribute
from efootprint.abstract_modeling_classes.object_linked_to_modeling_obj import ObjectLinkedToModelingObj
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject, AfterInitMeta
from efootprint.abstract_modeling_classes.modeling_update import ModelingUpdate


class ListLinkedToModelingObj(ObjectLinkedToModelingObj, list, metaclass=AfterInitMeta):
    def __init__(self, values=None):
        super().__init__()
        self.trigger_modeling_updates = False
        if values is not None:
            self.extend(values)

    def after_init(self):
        self.trigger_modeling_updates = True
    
    @staticmethod
    def check_value_type(value):
        if not isinstance(value, ModelingObject):
            raise ValueError(
                f"ListLinkedToModelingObjs only accept ModelingObjects as values, received {type(value)}")

    def set_modeling_obj_container(self, new_parent_modeling_object: ModelingObject, attr_name: str):
        super().set_modeling_obj_container(new_parent_modeling_object, attr_name)
        for value in self:
            value.set_modeling_obj_container(self.modeling_obj_container, self.attr_name_in_mod_obj_container)

    def __setitem__(self, index: int, value: ModelingObject):
        self.check_value_type(value)
        if self.trigger_modeling_updates:
            copied_list = list(self)
            copied_list[index] = value
            ModelingUpdate([[self.return_copy_with_same_attributes(), copied_list]])
            self.set_modeling_obj_container(None, None)

        value_to_set = ContextualModelingObjectAttribute(value)
        super().__setitem__(index, value_to_set)
        value_to_set.set_modeling_obj_container(self.modeling_obj_container, self.attr_name_in_mod_obj_container)

    def append(self, value: ModelingObject):
        self.check_value_type(value)
        if self.trigger_modeling_updates:
            copied_list = list(self)
            copied_list.append(value)
            ModelingUpdate([[self.return_copy_with_same_attributes(), copied_list]])
            self.set_modeling_obj_container(None, None)

        value_to_set = ContextualModelingObjectAttribute(value)
        super().append(value_to_set)
        value_to_set.set_modeling_obj_container(self.modeling_obj_container, self.attr_name_in_mod_obj_container)

    def to_json(self, with_calculated_attributes_data=False):
        return [elt.id for elt in self]

    def __repr__(self):
        return str(self.to_json())

    def __str__(self):
        return_str = "[\n"

        for item in self:
            return_str += f"{item}, \n"

        return_str = return_str + "]"

        return return_str

    def insert(self, index: int, value: ModelingObject):
        self.check_value_type(value)
        if self.trigger_modeling_updates:
            copied_list = list(self)
            copied_list.insert(index, value)
            ModelingUpdate([[self.return_copy_with_same_attributes(), copied_list]])
            self.set_modeling_obj_container(None, None)

        value_to_set = ContextualModelingObjectAttribute(value)
        super().insert(index, value_to_set)
        value_to_set.set_modeling_obj_container(self.modeling_obj_container, self.attr_name_in_mod_obj_container)

    def extend(self, values) -> None:
        if self.trigger_modeling_updates:
            copied_list = list(self)
            copied_list.extend(values)
            ModelingUpdate([[self.return_copy_with_same_attributes(), copied_list]])
            self.set_modeling_obj_container(None, None)

        initial_trigger_modeling_updates = copy(self.trigger_modeling_updates)
        self.trigger_modeling_updates = False
        for value in values:
            self.append(value)
        self.trigger_modeling_updates = initial_trigger_modeling_updates

    def pop(self, index: int = -1):
        if self.trigger_modeling_updates:
            copied_list = list(self)
            _ = copied_list.pop(index)
            ModelingUpdate([[self.return_copy_with_same_attributes(), copied_list]])
            self.set_modeling_obj_container(None, None)

        value = super().pop(index)
        value.set_modeling_obj_container(None, None)

        return value

    def remove(self, value: ContextualModelingObjectAttribute):
        if self.trigger_modeling_updates:
            copied_list = list(self)
            copied_list.remove(value)
            ModelingUpdate([[self.return_copy_with_same_attributes(), copied_list]])
            self.set_modeling_obj_container(None, None)

        super().remove(value)
        value.set_modeling_obj_container(None, None)

    def clear(self):
        if self.trigger_modeling_updates:
            ModelingUpdate([[self.return_copy_with_same_attributes(), []]])
            self.set_modeling_obj_container(None, None)

        for value in self:
            value.set_modeling_obj_container(None, None)
        super().clear()

    def __delitem__(self, index: int):
        if self.trigger_modeling_updates:
            copied_list = list(self)
            del copied_list[index]
            ModelingUpdate([[self.return_copy_with_same_attributes(), copied_list]])
            self.set_modeling_obj_container(None, None)

        value = self[index]
        value.set_modeling_obj_container(None, None)
        super().__delitem__(index)

    def __iadd__(self, values):
        self.extend(values)
        return self

    def __imul__(self, n: int):
        if self.trigger_modeling_updates:
            copied_list = list(self)
            copied_list *= n
            ModelingUpdate([[self.return_copy_with_same_attributes(), copied_list]])
            self.set_modeling_obj_container(None, None)

        initial_trigger_modeling_updates = copy(self.trigger_modeling_updates)
        self.trigger_modeling_updates = False
        for _ in range(n - 1):
            self.extend(self.copy())
        self.trigger_modeling_updates = initial_trigger_modeling_updates

        return self

    def __copy__(self):
        return ListLinkedToModelingObj([value for value in self])

    def return_copy_with_same_attributes(self):
        copied_list = ListLinkedToModelingObj(self)
        copied_list.set_modeling_obj_container(self.modeling_obj_container, self.attr_name_in_mod_obj_container)
        copied_list.trigger_modeling_updates = self.trigger_modeling_updates
        
        return copied_list
    


from typing import Type, Optional
from dataclasses import dataclass
import os

from efootprint.abstract_modeling_classes.object_linked_to_modeling_obj import ObjectLinkedToModelingObj
from efootprint.constants.units import u
from efootprint.logger import logger
from efootprint.utils.calculus_graph import build_calculus_graph
from efootprint.utils.graph_tools import add_unique_id_to_mynetwork


@dataclass
class Source:
    name: str
    link: Optional[str]


def retrieve_update_function_from_mod_obj_and_attr_name(mod_obj, attr_name):
    update_func_name = f"update_{attr_name}"
    update_func = getattr(mod_obj, update_func_name, None)

    if update_func is None:
        raise AttributeError(f"No update function associated to {attr_name} in {mod_obj.id}, please create it.")

    return update_func


def optimize_attr_updates_chain(attr_updates_chain):
    initial_chain_len = len(attr_updates_chain)
    attr_to_update_ids = [attr.id for attr in attr_updates_chain]
    optimized_chain = []

    for index in range(len(attr_updates_chain)):
        attr_to_update = attr_updates_chain[index]

        if attr_to_update.id not in attr_to_update_ids[index + 1:]:
            # Keep only last occurrence of each update function
            optimized_chain.append(attr_to_update)

    optimized_chain_len = len(optimized_chain)

    if optimized_chain_len != initial_chain_len:
        logger.info(f"Optimized update function chain from {initial_chain_len} to {optimized_chain_len} calculations")

    return optimized_chain


class ExplainableObject(ObjectLinkedToModelingObj):
    def __init__(
            self, value: object, label: str = None, left_parent: Type["ExplainableObject"] = None,
            right_parent: Type["ExplainableObject"] = None, operator: str = None, source: Source = None):
        super().__init__()
        self.simulation_twin = None
        self.baseline_twin = None
        self.simulation = None
        self.initial_modeling_obj_container = None
        self.value = value
        if not label and (left_parent is None and right_parent is None):
            raise ValueError(f"ExplainableObject without parent should have a label")
        self.source = source
        self.label = None
        self.set_label(label)
        self.left_parent = left_parent
        self.right_parent = right_parent
        self.operator = operator
        self.direct_ancestors_with_id = []
        self.direct_children_with_id = []

        for parent in (self.left_parent, self.right_parent):
            if parent is not None:
                self.direct_ancestors_with_id += [
                    ancestor_with_id for ancestor_with_id in parent.return_direct_ancestors_with_id_to_child()
                    if ancestor_with_id.id not in self.direct_ancestor_ids]

    def __copy__(self):
        cls = self.__class__
        new_instance = cls.__new__(cls)
        new_instance.__init__(value=self.value, label=self.label, source=getattr(self, "source", None))

        return new_instance

    def set_label(self, new_label):
        if self.source is not None and f"from {self.source.name}" not in new_label:
            self.label = f"{new_label} from {self.source.name}"
        else:
            self.label = new_label

        return self

    @property
    def has_parent(self):
        return self.left_parent is not None or self.right_parent is not None

    @property
    def direct_ancestor_ids(self):
        return [attr.id for attr in self.direct_ancestors_with_id]

    @property
    def direct_child_ids(self):
        return [attr.id for attr in self.direct_children_with_id]

    def replace_in_mod_obj_container_without_recomputation(self, new_value):
        value_to_set = new_value
        if isinstance(value_to_set, float) and self.dict_container is not None:
            from efootprint.abstract_modeling_classes.source_objects import SourceValue
            value_to_set = SourceValue(value_to_set * u.dimensionless)

        super().replace_in_mod_obj_container_without_recomputation(value_to_set)

    def set_modeling_obj_container(
            self, new_modeling_obj_container: Type["ModelingObject"] | None, attr_name: str | None):
        if not self.label:
            raise PermissionError(
                f"ExplainableObjects that are attributes of a ModelingObject should always have a label.")
        elif self.modeling_obj_container is not None and new_modeling_obj_container is not None\
                and self.modeling_obj_container != new_modeling_obj_container:
            raise PermissionError(
                f"{self} is already linked to {self.modeling_obj_container.id} and is trying to be linked to "
                f"{new_modeling_obj_container.id}.")

        if self.modeling_obj_container is not None:
            for direct_ancestor_with_id in self.direct_ancestors_with_id:
                direct_ancestor_with_id.remove_child_from_direct_children_with_id(direct_child=self)

        super().set_modeling_obj_container(new_modeling_obj_container, attr_name)

        if new_modeling_obj_container is not None:
            if self.initial_modeling_obj_container is None:
                self.initial_modeling_obj_container = new_modeling_obj_container
            for direct_ancestor_with_id in self.direct_ancestors_with_id:
                direct_ancestor_with_id.add_child_to_direct_children_with_id(direct_child=self)

    def return_direct_ancestors_with_id_to_child(self):
        if self.modeling_obj_container is not None:
            return [self]
        else:
            return [ancestor for ancestor in self.direct_ancestors_with_id
                    if ancestor.modeling_obj_container is not None]

    def add_child_to_direct_children_with_id(self, direct_child):
        if direct_child.id not in self.direct_child_ids:
            self.direct_children_with_id.append(direct_child)

    def remove_child_from_direct_children_with_id(self, direct_child):
        self.direct_children_with_id = [
            child for child in self.direct_children_with_id if child.id != direct_child.id]

    @property
    def all_descendants_with_id(self):
        all_descendants = []

        def retrieve_descendants(expl_obj: ExplainableObject, descendants_list):
            for child in expl_obj.direct_children_with_id:
                if child.id not in [elt.id for elt in descendants_list]:
                    descendants_list.append(child)
                retrieve_descendants(child, descendants_list)

        retrieve_descendants(self, all_descendants)

        return all_descendants

    @property
    def all_ancestors_with_id(self):
        all_ancestors = []

        def retrieve_ancestors(expl_obj: ExplainableObject, ancestors_list):
            for parent in expl_obj.direct_ancestors_with_id:
                if parent.id not in [elt.id for elt in ancestors_list]:
                    ancestors_list.append(parent)
                retrieve_ancestors(parent, ancestors_list)

        retrieve_ancestors(self, all_ancestors)

        return all_ancestors

    @property
    def attr_updates_chain(self):
        attr_updates_chain = []
        descendants = self.all_descendants_with_id
        has_been_added_to_chain_dict = {descendant.id: False for descendant in descendants if descendant.id != self.id}

        added_parents_with_children_to_add = [self]

        while len(added_parents_with_children_to_add) > 0:
            for added_parent in added_parents_with_children_to_add:
                drop_added_parent_from_list = True
                for child in added_parent.direct_children_with_id:
                    if not has_been_added_to_chain_dict[child.id]:
                        ancestors_that_belong_to_self_descendants = [
                            ancestor for ancestor in child.direct_ancestors_with_id
                            if ancestor.id in [ancestor.id for ancestor in descendants]]
                        if all([has_been_added_to_chain_dict[ancestor.id]
                                for ancestor in ancestors_that_belong_to_self_descendants]):
                            if not child.dict_container:
                                attr_updates_chain.append(child)
                            else:
                                attr_updates_chain.append(child.dict_container)
                            has_been_added_to_chain_dict[child.id] = True
                            if len(child.direct_children_with_id) > 0:
                                added_parents_with_children_to_add.append(child)
                        else:
                            # Wait for next iteration
                            drop_added_parent_from_list = False
                if drop_added_parent_from_list:
                    added_parents_with_children_to_add = [
                        child for child in added_parents_with_children_to_add
                        if child.id != added_parent.id]

        optimized_chain = optimize_attr_updates_chain(attr_updates_chain)

        return optimized_chain

    @property
    def update_function(self):
        if self.modeling_obj_container is None:
            raise ValueError(
                f"{self} doesn’t have a modeling_obj_container, hence it makes no sense "
                f"to look for its update function")
        update_func = retrieve_update_function_from_mod_obj_and_attr_name(
            self.modeling_obj_container, self.attr_name_in_mod_obj_container)

        return update_func

    @property
    def update_function_chain(self):
        return [attribute.update_function for attribute in self.attr_updates_chain]
    
    def generate_explainable_object_with_logical_dependency(
            self, explainable_condition: Type["ExplainableObject"]):
        return self.__class__(value=self.value, label=self.label, left_parent=self, right_parent=explainable_condition,
                              operator="logically dependent on")

    def explain(self, pretty_print=True):
        element_value_to_print = str(self)

        if self.left_parent is None and self.right_parent is None:
            return f"{self.label} = {element_value_to_print}"
        explain_tuples = self.compute_explain_nested_tuples()

        if pretty_print:
            return self.pretty_print_calculation(
                f"{self.label} = {self.print_tuple_element(explain_tuples, print_values_instead_of_labels=False)}"
                f" = {self.print_tuple_element(explain_tuples, print_values_instead_of_labels=True)}"
                f" = {element_value_to_print}")
        else:
            return f"{self.label} = {self.print_tuple_element(explain_tuples, print_values_instead_of_labels=False)}" \
                f" = {self.print_tuple_element(explain_tuples, print_values_instead_of_labels=True)}" \
                f" = {element_value_to_print}"

    def compute_explain_nested_tuples(self, return_label_if_self_has_one=False):
        if return_label_if_self_has_one and self.label:
            return self

        left_explanation = None
        right_explanation = None

        if self.left_parent:
            left_explanation = self.left_parent.compute_explain_nested_tuples(return_label_if_self_has_one=True)
        if self.right_parent:
            right_explanation = self.right_parent.compute_explain_nested_tuples(return_label_if_self_has_one=True)

        if left_explanation is None and right_explanation is None:
            raise ValueError("Object to explain should have at least one child")

        return left_explanation, self.operator, right_explanation

    def print_tuple_element(self, tuple_element: object, print_values_instead_of_labels: bool):
        if isinstance(tuple_element, ExplainableObject):
            if print_values_instead_of_labels:
                return str(tuple_element)
            else:
                return f"{tuple_element.label}"
        elif type(tuple_element) == str:
            return tuple_element
        elif type(tuple_element) == tuple:
            if tuple_element[1] is None:
                return f"{self.print_tuple_element(tuple_element[0], print_values_instead_of_labels)}"
            if tuple_element[2] is None:
                return f"{tuple_element[1]}" \
                       f" of ({self.print_tuple_element(tuple_element[0], print_values_instead_of_labels)})"

            left_parenthesis = False
            right_parenthesis = False

            if tuple_element[1] == "/":
                if type(tuple_element[2]) == tuple:
                    right_parenthesis = True
                if type(tuple_element[0]) == tuple and tuple_element[0][1] != "*":
                    left_parenthesis = True
            elif tuple_element[1] == "*":
                if type(tuple_element[0]) == tuple and tuple_element[0][1] != "*":
                    left_parenthesis = True
                if type(tuple_element[2]) == tuple and tuple_element[2][1] != "*":
                    right_parenthesis = True
            elif tuple_element[1] == "-":
                if type(tuple_element[2]) == tuple and tuple_element[2][1] in ["+", "-"]:
                    right_parenthesis = True
            elif tuple_element[1] == "+":
                pass

            lp_open = ""
            lp_close = ""
            rp_open = ""
            rp_close = ""

            if left_parenthesis:
                lp_open = "("
                lp_close = ")"
            if right_parenthesis:
                rp_open = "("
                rp_close = ")"

            return f"{lp_open}{self.print_tuple_element(tuple_element[0], print_values_instead_of_labels)}{lp_close}" \
                   f" {tuple_element[1]}" \
                   f" {rp_open}{self.print_tuple_element(tuple_element[2], print_values_instead_of_labels)}{rp_close}"

    @staticmethod
    def pretty_print_calculation(calc_str):
        return calc_str.replace(" = ", "\n=\n")

    def calculus_graph_to_file(
            self, filename=None, colors_dict=None, x_multiplier=150, y_multiplier=150, width="1800px", height="900px",
            notebook=False, max_depth=100):
        if colors_dict is None:
            colors_dict = {"user data": "gold", "default": "darkred"}
        calculus_graph = build_calculus_graph(
            self, colors_dict, x_multiplier, y_multiplier, width, height, notebook, max_depth=max_depth)

        if filename is None:
            filename = os.path.join(".", f"{self.label} calculus graph.html")

        calculus_graph.show(filename, notebook=notebook)

        add_unique_id_to_mynetwork(filename)

        if notebook:
            from IPython.display import HTML

            return HTML(filename)

    def to_json(self, with_calculated_attributes_data=False):
        output_dict = {"label": self.label}

        if isinstance(self.value, str):  # Case of technology in WebApplication
            output_dict["value"] = self.value
        elif getattr(self.value, "zone", None) is not None:  # Case of timezone in Country class
            output_dict["zone"] = self.value.zone

        if self.source is not None:
            output_dict["source"] = {"name": self.source.name, "link": self.source.link}

        if with_calculated_attributes_data:
            output_dict["id"] = self.id
            output_dict["direct_ancestors_with_id"] = [elt.id for elt in self.direct_ancestors_with_id]
            output_dict["direct_children_with_id"] = [elt.id for elt in self.direct_children_with_id]

        return output_dict

    def __repr__(self):
        return str(self)

    def __str__(self):
        self_as_str = str(self.value)
        if len(self_as_str) > 60:
            self_as_str = self_as_str[:60] + "..."

        return self_as_str

    def __eq__(self, other):
        if isinstance(other, ExplainableObject):
            return self.value == other.value
        else:
            return False

    def __hash__(self):
        return hash(self.value)



from pandas import DataFrame


def plot_baseline_and_simulation_dfs(baseline_df: DataFrame, simulated_values_df: DataFrame=None,
                                     figsize=(10, 4), xlims=None):
    from matplotlib import pyplot as plt
    import matplotlib.dates as mdates

    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=figsize)

    if xlims is not None:
        start, end = xlims
        filtered_df = baseline_df[(baseline_df.index.to_timestamp() >= start)
                                  & (baseline_df.index.to_timestamp() <= end)]
        ax.set_xlim(xlims)
        max_val = filtered_df["value"].max().magnitude
        min_val = filtered_df["value"].min().magnitude
        if simulated_values_df is not None:
            simulated_filtered_df = simulated_values_df[(simulated_values_df.index.to_timestamp() >= start)
                                                       & (simulated_values_df.index.to_timestamp() <= end)]
            max_val = max(max_val, simulated_filtered_df["value"].max().magnitude)
            min_val = min(min_val, simulated_filtered_df["value"].min().magnitude)
        offset = (max_val - min_val) * 0.1
        ax.set_ylim([min_val - offset, max_val + offset])

    ax.plot(baseline_df.index.to_timestamp().values, baseline_df["value"].values.data, label="baseline")

    if simulated_values_df is not None:
            ax.plot(simulated_values_df.index.to_timestamp().values,
                    simulated_values_df["value"].values.data, label="simulated")
        
    ax.legend()
    plt.ylabel(f"{baseline_df.dtypes.value.units:~}")

    locator = mdates.AutoDateLocator(minticks=3, maxticks=12)
    formatter = mdates.ConciseDateFormatter(locator)
    ax.xaxis.set_major_locator(locator)
    ax.xaxis.set_major_formatter(formatter)

    return ax


from typing import List

from efootprint.abstract_modeling_classes.explainable_objects import EmptyExplainableObject, ExplainableQuantity
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.sources import Sources
from efootprint.constants.units import u


class Network(ModelingObject):
    @classmethod
    def default_values(cls):
        return {
            "bandwidth_energy_intensity": SourceValue(0.1 * u.kWh / u.GB)
        }

    @classmethod
    def wifi_network(cls, name: str = "Default wifi network"):
        return cls(
            name=name, bandwidth_energy_intensity=SourceValue(0.05 * u("kWh/GB"), Sources.TRAFICOM_STUDY))

    @classmethod
    def mobile_network(cls, name: str = "Default mobile network"):
        return cls(
            name=name, bandwidth_energy_intensity=SourceValue(0.12 * u("kWh/GB"), Sources.TRAFICOM_STUDY))

    @classmethod
    def archetypes(cls):
        return [cls.wifi_network, cls.mobile_network]

    def __init__(self, name: str, bandwidth_energy_intensity: ExplainableQuantity):
        super().__init__(name)
        self.energy_footprint = EmptyExplainableObject()
        self.bandwidth_energy_intensity = bandwidth_energy_intensity.set_label(
            f"bandwith energy intensity of {self.name}")

    @property
    def calculated_attributes(self):
        return ["energy_footprint"]

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List:
        return []

    @property
    def usage_patterns(self):
        return self.modeling_obj_containers

    @property
    def systems(self) -> List:
        return list(set(sum([up.systems for up in self.usage_patterns], start=[])))

    @property
    def jobs(self) -> List[ModelingObject]:
        return list(set(sum([up.jobs for up in self.usage_patterns], start=[])))

    def update_energy_footprint(self):
        hourly_data_transferred_per_up = {up: EmptyExplainableObject() for up in self.usage_patterns}
        for job in self.jobs:
            job_ups_in_network_ups = [up for up in job.usage_patterns if up in self.usage_patterns]
            for up in job_ups_in_network_ups:
                hourly_data_transferred_per_up[up] += job.hourly_data_transferred_per_usage_pattern[up]

        energy_footprint = EmptyExplainableObject()
        for up in self.usage_patterns:
            up_network_consumption = (
                        self.bandwidth_energy_intensity * hourly_data_transferred_per_up[up]).to(u.kWh).set_label(
                f"{up.name} network energy consumption")

            energy_footprint += up_network_consumption * up.country.average_carbon_intensity

        self.energy_footprint = energy_footprint.to(u.kg).set_label(f"Hourly {self.name} energy footprint")



from copy import copy
from datetime import datetime
from typing import List

import pandas as pd

from efootprint.abstract_modeling_classes.contextual_modeling_object_attribute import ContextualModelingObjectAttribute
from efootprint.abstract_modeling_classes.explainable_object_base_class import ExplainableObject, \
    optimize_attr_updates_chain
from efootprint.abstract_modeling_classes.object_linked_to_modeling_obj import ObjectLinkedToModelingObj
from efootprint.abstract_modeling_classes.explainable_objects import ExplainableHourlyQuantities, EmptyExplainableObject
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject, optimize_mod_objs_computation_chain
from efootprint.logger import logger


def compute_attr_updates_chain_from_mod_objs_computation_chain(mod_objs_computation_chain: List[ModelingObject]):
    attr_updates_chain = []
    for mod_obj in mod_objs_computation_chain:
        for calculated_attribute in mod_obj.calculated_attributes:
            attr_updates_chain.append(getattr(mod_obj, calculated_attribute))

    return attr_updates_chain


class ModelingUpdate:
    def __init__(
            self, changes_list: List[List[ObjectLinkedToModelingObj | list | dict]], simulation_date: datetime = None):
        self.updated_values_set = False
        first_changed_val = changes_list[0][0]
        self.system = None
        if isinstance(first_changed_val, ObjectLinkedToModelingObj):
            if first_changed_val.modeling_obj_container.systems:
                self.system = first_changed_val.modeling_obj_container.systems[0]
        else:
            raise ValueError(
                f"First changed value {first_changed_val} is not an ObjectLinkedToModelingObj")
        self.changes_list = changes_list
        self.parse_changes_list()
        if self.changes_list and self.system:
            self.system.previous_total_energy_footprints_sum_over_period = (
                self.system.total_energy_footprint_sum_over_period)
            self.system.previous_total_fabrication_footprints_sum_over_period = \
                self.system.total_fabrication_footprint_sum_over_period
            self.system.previous_change = changes_list
            self.system.all_changes += changes_list

        self.simulation_date = simulation_date
        self.simulation_date_as_hourly_freq = None
        if simulation_date is not None and self.system is not None:
            self.system.simulation = self
            self.simulation_date_as_hourly_freq = pd.Timestamp(simulation_date).to_period(freq="h")
        
        self.mod_objs_computation_chain = self.compute_mod_objs_computation_chain()
        if self.mod_objs_computation_chain:
            logger.info(f"{len(self.mod_objs_computation_chain)} recomputed objects: "
                        f"{[mod_obj.id for mod_obj in self.mod_objs_computation_chain]}")
        self.attr_updates_chain_from_mod_objs_computation_chains = (
            compute_attr_updates_chain_from_mod_objs_computation_chain(self.mod_objs_computation_chain))
        self.values_to_recompute = self.generate_optimized_attr_updates_chain()

        self.ancestors_not_in_computation_chain = []
        self.hourly_quantities_to_filter = []
        self.filtered_hourly_quantities = []
        self.ancestors_to_replace_by_copies = []
        self.replaced_ancestors_copies = []
        if self.simulation_date is not None:
            self.make_simulation_specific_operations()

        self.apply_changes()
        for new_sourcevalue in self.new_sourcevalues:
            mod_obj_container = new_sourcevalue.modeling_obj_container
            mod_obj_container.check_belonging_to_authorized_values(
                new_sourcevalue.attr_name_in_mod_obj_container, new_sourcevalue, mod_obj_container.list_values(),
                mod_obj_container.conditional_list_values(), mod_obj_container.attributes_with_depending_values())
        self.recomputed_values = self.recompute_attributes()
        self.updated_values_set = True

        if self.simulation_date is not None:
            self.link_simulated_and_baseline_twins()

        self.all_previous_obj_linked_to_mod_obj = (
                [change[0] for change in self.changes_list] + self.hourly_quantities_to_filter
                + self.ancestors_to_replace_by_copies + self.values_to_recompute)
        self.all_new_obj_linked_to_mod_obj = (
                [change[1] for change in self.changes_list] + self.filtered_hourly_quantities
                + self.replaced_ancestors_copies + self.recomputed_values)

        if simulation_date is not None:
            self.reset_values()

    def parse_changes_list(self):
        indexes_to_skip = []
        for index in range(len(self.changes_list)):
            old_value, new_value = self.changes_list[index]
            assert isinstance(old_value, ObjectLinkedToModelingObj)
            if new_value is None:
                assert isinstance(old_value, ExplainableObject)
                self.changes_list[index][1] = EmptyExplainableObject()
            else:
                mod_obj_container = old_value.modeling_obj_container
                mod_obj_container.check_input_value_type_positivity_and_unit(
                    old_value.attr_name_in_mod_obj_container, new_value, mod_obj_container.default_values())

            if isinstance(new_value, list):
                from efootprint.abstract_modeling_classes.list_linked_to_modeling_obj import ListLinkedToModelingObj
                self.changes_list[index][1] = ListLinkedToModelingObj(new_value)
            if isinstance(new_value, ModelingObject):
                self.changes_list[index][1] = ContextualModelingObjectAttribute(new_value)

            if not isinstance(self.changes_list[index][1], ObjectLinkedToModelingObj):
                raise ValueError(
                    f"New e-footprint attributes should be ObjectLinkedToModelingObj,"
                    f" got {old_value} of type {type(old_value)} trying to be set to an object "
                    f"of type {type(new_value)}")

            if old_value == new_value:
                logger.warning(
                    f"{old_value.id} is updated to itself. "
                    f"It happens when using my_mod_obj.list_attribute += other list syntax. "
                    f"Otherwise This is surprising, you might want to double check your action. "
                    f"The link update logic will be skipped.")
                indexes_to_skip.append(index)

        for index in sorted(indexes_to_skip, reverse=True):
            del self.changes_list[index]

    def compute_mod_objs_computation_chain(self):
        from efootprint.abstract_modeling_classes.list_linked_to_modeling_obj import ListLinkedToModelingObj
        mod_objs_computation_chain = []
        for old_value, new_value in self.changes_list:
            if isinstance(old_value, ContextualModelingObjectAttribute):
                mod_objs_computation_chain += (
                    old_value.modeling_obj_container.compute_mod_objs_computation_chain_from_old_and_new_modeling_objs(
                        old_value, new_value, optimize_chain=False))
            elif isinstance(old_value, ListLinkedToModelingObj):
                mod_objs_computation_chain += (
                    old_value.modeling_obj_container.compute_mod_objs_computation_chain_from_old_and_new_lists(
                        old_value, new_value, optimize_chain=False))

        optimized_chain = optimize_mod_objs_computation_chain(mod_objs_computation_chain)

        return optimized_chain

    def apply_changes(self):
        for old_value, new_value in self.changes_list:
            old_value.replace_in_mod_obj_container_without_recomputation(new_value)

    def make_simulation_specific_operations(self):
        assert self.simulation_date is not None
        self.ancestors_not_in_computation_chain = self.compute_ancestors_not_in_computation_chain()
        self.hourly_quantities_to_filter = self.compute_hourly_quantities_to_filter()
        if self.mod_objs_computation_chain:
            # The simulation will change the calculation graph, so we need to replace all ancestors not in
            # computation chain by their copies to keep the original calculation graph unchanged
            self.ancestors_to_replace_by_copies = [
                ancestor for ancestor in self.ancestors_not_in_computation_chain
                if ancestor.id not in [value.id for value in self.hourly_quantities_to_filter]]
            self.replaced_ancestors_copies = self.replace_ancestors_not_in_computation_chain_by_copies()
        self.filter_hourly_quantities_to_filter()

    def recompute_attributes(self):
        recomputed_values = []
        for value_to_recompute in self.values_to_recompute:
            attr_name_in_mod_obj_container = value_to_recompute.attr_name_in_mod_obj_container
            modeling_obj_container = value_to_recompute.modeling_obj_container
            logger.debug(f"Recomputing {attr_name_in_mod_obj_container} in {modeling_obj_container.id}")
            value_to_recompute.update_function()
            recomputed_value = getattr(modeling_obj_container, attr_name_in_mod_obj_container)
            recomputed_values.append(recomputed_value)

        return recomputed_values

    @property
    def old_sourcevalues(self):
        return [old_value for old_value, new_value in self.changes_list if isinstance(old_value, ExplainableObject)]

    @property
    def new_sourcevalues(self):
        return [new_value for old_value, new_value in self.changes_list if isinstance(old_value, ExplainableObject)]

    def generate_optimized_attr_updates_chain(self):
        attr_updates_chain_from_attributes_updates = sum(
            [old_value.attr_updates_chain for old_value in self.old_sourcevalues], start=[])

        optimized_chain = optimize_attr_updates_chain(
            self.attr_updates_chain_from_mod_objs_computation_chains + attr_updates_chain_from_attributes_updates)

        optimized_chain_without_previous_nor_initial_values = [
            attr for attr in optimized_chain if not attr.attr_name_in_mod_obj_container.startswith("previous_")
                                                and not attr.attr_name_in_mod_obj_container.startswith("initial_")]

        return optimized_chain_without_previous_nor_initial_values

    def compute_ancestors_not_in_computation_chain(self):
        all_ancestors_of_values_to_recompute = sum(
            [value.all_ancestors_with_id for value in self.values_to_recompute], start=[])
        deduplicated_all_ancestors_of_values_to_recompute = []
        for ancestor in all_ancestors_of_values_to_recompute:
            if ancestor.id not in [elt.id for elt in deduplicated_all_ancestors_of_values_to_recompute]:
                deduplicated_all_ancestors_of_values_to_recompute.append(ancestor)
        values_to_recompute_ids = [elt.id for elt in self.values_to_recompute]
        old_sourcevalues_ids = [old_value.id for old_value in self.old_sourcevalues]
        ancestors_not_in_computation_chain = [
            ancestor for ancestor in deduplicated_all_ancestors_of_values_to_recompute
            if ancestor.id not in values_to_recompute_ids + old_sourcevalues_ids]

        return ancestors_not_in_computation_chain

    def compute_hourly_quantities_to_filter(self):
        hourly_quantities_ancestors_not_in_computation_chain = [
            ancestor for ancestor in self.ancestors_not_in_computation_chain
            if isinstance(ancestor, ExplainableHourlyQuantities)]
        hourly_quantities_to_filter = []

        global_min_date = None
        global_max_date = None

        for ancestor in hourly_quantities_ancestors_not_in_computation_chain:
            min_date = ancestor.value.index.min()
            max_date = ancestor.value.index.max()
            if global_min_date is None:
                global_min_date = min_date
            if global_max_date is None:
                global_max_date = max_date
            global_min_date = min(min_date, global_min_date)
            global_max_date = max(max_date, global_max_date)
            if self.simulation_date_as_hourly_freq <= max_date:
                hourly_quantities_to_filter.append(ancestor)

        if not (global_min_date <= self.simulation_date_as_hourly_freq <= global_max_date):
            raise ValueError(
                f"Can’t start a simulation on the {self.simulation_date_as_hourly_freq} because "
                f"{self.simulation_date_as_hourly_freq}doesn’t belong to the existing modeling period "
                f"{global_min_date} to {global_max_date}")

        return hourly_quantities_to_filter

    def filter_hourly_quantities_to_filter(self):
        for hourly_quantities in self.hourly_quantities_to_filter:

            new_value = ExplainableHourlyQuantities(
                hourly_quantities.value[hourly_quantities.value.index >= self.simulation_date_as_hourly_freq],
                hourly_quantities.label, hourly_quantities.left_parent, hourly_quantities.right_parent,
                hourly_quantities.operator, hourly_quantities.source
            )
            if len(new_value) == 0:
                new_value = EmptyExplainableObject()
            hourly_quantities.replace_in_mod_obj_container_without_recomputation(new_value)
            self.filtered_hourly_quantities.append(new_value)

    def replace_ancestors_not_in_computation_chain_by_copies(self):
        copies = []
        for ancestor_to_replace_by_copy in self.ancestors_to_replace_by_copies:
            # Replace all ancestors not in computation chain by their copy so that the original calculation graph
            # will remain unchanged when the simulation is over
            ancestor_copy = copy(ancestor_to_replace_by_copy)
            ancestor_to_replace_by_copy.replace_in_mod_obj_container_without_recomputation(ancestor_copy)
            copies.append(ancestor_copy)

        return copies

    def reset_values(self):
        if self.updated_values_set:
            for new_value, previous_value in zip(
                    self.all_new_obj_linked_to_mod_obj, self.all_previous_obj_linked_to_mod_obj):
                new_value.replace_in_mod_obj_container_without_recomputation(previous_value)
            self.updated_values_set = False

    def set_updated_values(self):
        if not self.updated_values_set:
            for new_value, previous_value in zip(
                    self.all_new_obj_linked_to_mod_obj, self.all_previous_obj_linked_to_mod_obj):
                previous_value.replace_in_mod_obj_container_without_recomputation(new_value)
            self.updated_values_set = True

    def link_simulated_and_baseline_twins(self):
        assert self.simulation_date is not None
        for value_to_recompute, recomputed_value in zip(self.values_to_recompute, self.recomputed_values):
            value_to_recompute.simulation_twin = recomputed_value
            recomputed_value.baseline_twin = value_to_recompute
            value_to_recompute.simulation = self
            recomputed_value.simulation = self



from copy import copy

import pytz

from efootprint.constants.units import u
from efootprint.constants.sources import Source,Sources
from efootprint.abstract_modeling_classes.source_objects import SourceValue, SourceObject
from efootprint.core.country import Country


def tz(timezone: str):
    return SourceObject(pytz.timezone(timezone), Sources.USER_DATA)


def country_generator(country_name, country_short_name, country_avg_carbon_int, timezone):
    def return_country():
        return Country(country_name, country_short_name, copy(country_avg_carbon_int), copy(timezone))

    return return_country


class Countries:
    # TODO: Add other countries and automate data retrieval
    source = Source("Our world in data", "https://ourworldindata.org/energy#country-profiles")
    FRANCE = country_generator("France", "FRA", SourceValue(85 * u.g / u.kWh, source), tz('Europe/Paris'))
    # EUROPE = country_generator("Europe", "EUR", SourceValue(278 * u.g / u.kWh, source), None)
    GERMANY = country_generator("Germany", "DEU", SourceValue(386 * u.g / u.kWh, source), tz('Europe/Berlin'))
    FINLAND = country_generator("Finland", "FIN", SourceValue(132 * u.g / u.kWh, source), tz('Europe/Helsinki'))
    AUSTRIA = country_generator("Austria", "AUT", SourceValue(158 * u.g / u.kWh, source), tz('Europe/Vienna'))
    POLAND = country_generator("Poland", "POL", SourceValue(635 * u.g / u.kWh, source), tz('Europe/Warsaw'))
    NORWAY = country_generator("Norway", "NOR", SourceValue(26 * u.g / u.kWh, source), tz('Europe/Oslo'))
    HUNGARY = country_generator("Hungary", "HUN", SourceValue(223 * u.g / u.kWh, source), tz('Europe/Budapest'))
    UNITED_KINGDOM = country_generator("United Kingdom", "GBR", SourceValue(268 * u.g / u.kWh, source), tz('Europe/London'))
    BELGIUM = country_generator("Belgium", "BEL", SourceValue(165 * u.g / u.kWh, source), tz('Europe/Brussels'))
    ITALY = country_generator("Italy", "IT", SourceValue(371 * u.g / u.kWh, source), tz('Europe/Rome'))
    ROMANIA = country_generator("Romania", "RO", SourceValue(264 * u.g / u.kWh, source), tz('Europe/Bucharest'))
    MALAYSIA = country_generator("Malaysia", "MY", SourceValue(549 * u.g / u.kWh, source), tz('Asia/Kuala_Lumpur'))
    MOROCCO = country_generator("Morocco", "MA", SourceValue(610 * u.g / u.kWh, source), tz('Africa/Casablanca'))
    TUNISIA = country_generator("Tunisia", "TN", SourceValue(468 * u.g / u.kWh, source), tz('Africa/Tunis'))
    ALGERIA = country_generator("Algeria", "DZ", SourceValue(488 * u.g / u.kWh, source), tz('Africa/Algiers'))
    SENEGAL = country_generator("Senegal", "SN", SourceValue(503 * u.g / u.kWh, source), tz('Africa/Dakar'))
    # UNITED_STATES = country_generator("United States", "US", SourceValue(379 * u.g / u.kWh, source), None)
    # BRAZIL = country_generator("Brazil", "BR", SourceValue(159 * u.g / u.kWh, source), None)



import pint
from pint import UnitRegistry

from efootprint.constants.files import CUSTOM_UNITS_PATH

u = UnitRegistry(cache_folder=":auto:")
u.load_definitions(CUSTOM_UNITS_PATH)
u.default_locale = 'en_EN'
pint.set_application_registry(u)

from pint_pandas import PintType


u = PintType.ureg



from typing import Dict

from efootprint.abstract_modeling_classes.contextual_modeling_object_attribute import ContextualModelingObjectAttribute
from efootprint.abstract_modeling_classes.object_linked_to_modeling_obj import ObjectLinkedToModelingObj
from efootprint.abstract_modeling_classes.explainable_objects import EmptyExplainableObject
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject, ABCAfterInitMeta
from efootprint.abstract_modeling_classes.modeling_update import ModelingUpdate
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u


class WeightedModelingObjectsDict(ObjectLinkedToModelingObj, dict, metaclass=ABCAfterInitMeta):
    def __init__(self, weighted_modeling_objects_dict: Dict[ModelingObject, float | SourceValue]):
        super().__init__()
        self.trigger_modeling_updates = False
        self.object_type = None
        for key, value in weighted_modeling_objects_dict.items():
            self[key] = value

    def after_init(self):
        self.trigger_modeling_updates = True

    def compute_weighted_attr_sum(self, attr_name: str):
        weighted_sum = EmptyExplainableObject()
        for modeling_obj, modeling_obj_weight in self.items():
            weighted_sum += getattr(modeling_obj, attr_name) * modeling_obj_weight

        return weighted_sum.set_label(
            f"Weighted sum of {attr_name} over {[modeling_obj.id for modeling_obj in self.keys()]}")

    def set_modeling_obj_container(self, new_parent_modeling_object: ModelingObject, attr_name: str):
        super().set_modeling_obj_container(new_parent_modeling_object, attr_name)

        for key, value in self.items():
            key.set_modeling_obj_container(new_parent_modeling_object, attr_name)
            value.set_modeling_obj_container(new_parent_modeling_object, attr_name)

    def __setitem__(self, key, value: SourceValue):
        assert isinstance(key, ModelingObject)
        if self.object_type is None:
            assert type(key) == self.object_type
        else:
            self.object_type = type(key)
        contextual_modeling_object_attribute_key = ContextualModelingObjectAttribute(
            key, self.modeling_obj_container, self.attr_name_in_mod_obj_container)

        assert isinstance(value, float) or isinstance(value, SourceValue)
        value_to_set = value
        if isinstance(value, SourceValue):
            assert value.value.check("[]")
        else:
            value_to_set = SourceValue(value * u.dimensionless)

        if not self.trigger_modeling_updates:
            super().__setitem__(contextual_modeling_object_attribute_key, value_to_set)
            value_to_set.set_modeling_obj_container(
                new_modeling_obj_container=self.modeling_obj_container, attr_name=self.attr_name_in_mod_obj_container)
        else:
            if key in self.keys():
                ModelingUpdate([[self[key], value_to_set]])
            else:
                copied_dict = dict(self)
                copied_dict[contextual_modeling_object_attribute_key] = value_to_set
                ModelingUpdate([[self, copied_dict]])
                
    def to_json(self, with_calculated_attributes_data=False):
        output_dict = {}

        for key, value in self.items():
            output_dict[key.id] = value.to_json(with_calculated_attributes_data)

        return output_dict

    def __repr__(self):
        return str(self.to_json())

    def __str__(self):
        return_str = "{\n"

        for key, value in self.items():
            return_str += f"{key.id}: {value}, \n"

        return_str = return_str + "}"

        return return_str

    def __delitem__(self, key):
        self[key].set_modeling_obj_container(None, None)
        if self.trigger_modeling_updates:
            copied_dict = dict(self)
            copied_dict.__delitem__(key)
            ModelingUpdate([[self, copied_dict]])
        else:
            super().__delitem__(key)

    def pop(self, key):
        if self.trigger_modeling_updates:
            copied_dict = dict(self)
            value = copied_dict.pop(key)
            ModelingUpdate([[self, copied_dict]])
        else:
            value = super().pop(key)

        value.set_modeling_obj_container(None, None)
        
        return value

    def popitem(self):
        if self.trigger_modeling_updates:
            copied_dict = dict(self)
            key, value = copied_dict.popitem()
            ModelingUpdate([[self, copied_dict]])
        else:
            key, value = super().popitem()

        value.set_modeling_obj_container(None, None)
        
        return key, value

    def clear(self):
        for key, value in self.items():
            value.set_modeling_obj_container(None, None)
        
        if self.trigger_modeling_updates:
            ModelingUpdate([[self, {}]])
        else:
            super().clear()
        
    def update(self, __m, **kwargs):
        if self.trigger_modeling_updates:
            copied_dict = dict(self)
            copied_dict.update(__m, **kwargs)
            # TODO: move this to ModelingUpdate
            value_updates = []
            new_dict_part = __m.update(kwargs)
            for key, value in new_dict_part.items():
                if key in self.keys():
                    value_updates.append((self[key], value))
            ModelingUpdate(value_updates + [(self, copied_dict)])
        else:
            super().update(__m, **kwargs)

    def copy(self):
        raise NotImplementedError("WeightedModelingObjectsDict cannot be copied")

    def fromkeys(cls, __iterable, __value = None):
        dict_from_keys = dict.fromkeys(__iterable, __value)

        return WeightedModelingObjectsDict(dict_from_keys)

    def __eq__(self, other):
        raise NotImplementedError("WeightedModelingObjectsDict cannot be compared for equality")



from datetime import datetime, timedelta
from typing import List

import numpy as np
import pandas as pd
import pint

from efootprint.constants.units import u
from efootprint.abstract_modeling_classes.source_objects import SourceHourlyValues


def create_random_hourly_usage_df(
        timespan: pint.Quantity = 1 * u.day, min_val: int = 1, max_val: int = 10,
        start_date: datetime = datetime.strptime("2025-01-01", "%Y-%m-%d"),
        pint_unit: pint.Unit = u.dimensionless):
    nb_days = timespan.to(u.day).magnitude
    end_date = start_date + timedelta(days=nb_days)
    period_index = pd.period_range(start=start_date, end=end_date, freq='h')

    data = np.random.randint(min_val, max_val, size=len(period_index))
    df = pd.DataFrame(data, index=period_index, columns=['value'], dtype=f"pint[{str(pint_unit)}]")

    return df


def create_hourly_usage_df_from_list(
        input_list: List[float], start_date: datetime = datetime.strptime("2025-01-01", "%Y-%m-%d"),
        pint_unit: pint.Unit = u.dimensionless):
    end_date = start_date + timedelta(hours=len(input_list) - 1)
    period_index = pd.period_range(start=start_date, end=end_date, freq='h')

    df = pd.DataFrame(input_list, index=period_index, columns=['value'], dtype=f"pint[{str(pint_unit)}]")

    return df


def create_source_hourly_values_from_list(
    input_list: List[float], start_date: datetime = datetime.strptime("2025-01-01", "%Y-%m-%d"),
        pint_unit: pint.Unit = u.dimensionless):
    df = create_hourly_usage_df_from_list(input_list, start_date, pint_unit)

    return SourceHourlyValues(df)


def linear_growth_hourly_values(
        timespan: pint.Quantity, start_value: int, end_value: int,
        start_date: datetime = datetime.strptime("2025-01-01", "%Y-%m-%d"),
        pint_unit: pint.Unit = u.dimensionless):
    nb_of_hours = int(timespan.to(u.hour).magnitude)
    linear_growth = np.linspace(start_value, end_value, nb_of_hours)

    df = create_hourly_usage_df_from_list(linear_growth, start_date, pint_unit)

    return SourceHourlyValues(df)


def sinusoidal_fluct_hourly_values(
        timespan: pint.Quantity, sin_fluct_amplitude: int, sin_fluct_period_in_hours: int,
        start_date: datetime = datetime.strptime("2025-01-01", "%Y-%m-%d"),
        pint_unit: pint.Unit = u.dimensionless):
    nb_of_hours = int(timespan.to(u.hour).magnitude)
    time = np.arange(nb_of_hours)
    sinusoidal_fluctuation = sin_fluct_amplitude * np.sin(2 * np.pi * time / sin_fluct_period_in_hours)

    df = create_hourly_usage_df_from_list(sinusoidal_fluctuation, start_date, pint_unit)

    return SourceHourlyValues(df)


def daily_fluct_hourly_values(
        timespan: pint.Quantity, fluct_scale: float, hour_of_day_for_min_value: int = 4,
        start_date: datetime = datetime.strptime("2025-01-01", "%Y-%m-%d"),
        pint_unit: pint.Unit = u.dimensionless):
    assert fluct_scale > 0
    assert fluct_scale <= 1
    nb_of_hours = int(timespan.to(u.hour).magnitude)
    time = np.arange(nb_of_hours)
    hour_of_day = [(start_date.hour + x) % 24 for x in time]

    daily_fluctuation = (
            np.full(shape=len(hour_of_day), fill_value=1)
            + fluct_scale * np.sin(
                (3 * np.pi / 2)
                + (2 * np.pi
                    * (hour_of_day - np.full(shape=len(hour_of_day), fill_value=hour_of_day_for_min_value, dtype=int))
                    / 24
                )
            )
        )

    df = create_hourly_usage_df_from_list(daily_fluctuation, start_date, pint_unit)

    return SourceHourlyValues(df)


def create_hourly_usage_from_frequency(
        timespan: pint.Quantity, input_volume: float, frequency: str, active_days: list = None,
        hours: list = None, start_date: datetime = datetime.strptime("2025-01-01", "%Y-%m-%d"),
        pint_unit: pint.Unit = u.dimensionless):
    if frequency not in ['daily', 'weekly', 'monthly', 'yearly']:
        raise ValueError(f"frequency must be one of 'daily', 'weekly', 'monthly', or 'yearly', got {frequency}.")

    if frequency == 'daily' and active_days is not None:
        raise ValueError(f"active_days must be None for daily frequency, got {active_days}.")

    end_date = start_date + timedelta(days=timespan.to(u.day).magnitude)

    if active_days is None:
        if frequency == "weekly":
            active_days = [0]  # default to midnight or Monday
        else:
            active_days = [1]  # default to first day of month or first day of year

    if hours is None:
        hours = [0]  # default to midnight

    period_index = pd.period_range(start=start_date, end=end_date, freq='h')
    # Important to have fill_value be 0.0 otherwise values will be cast to int
    values = np.full(shape=len(period_index), fill_value=0.0)

    for i, period in enumerate(period_index):
        hour_of_day = period.hour  # Hour of the day, 0 to 23
        day_of_week = period.day_of_week  # Day of the week, 0 to 6
        day_of_month = period.day  # Day of the month, 1 to 31
        day_of_year = period.day_of_year  # Day of the year, 1 to 365/366
        if frequency == 'daily':
            if hour_of_day in hours:
                values[i] = input_volume
        elif frequency == 'weekly':
            if day_of_week in active_days and hour_of_day in hours:
                values[i] = input_volume
        elif frequency == 'monthly':
            if day_of_month in active_days and hour_of_day in hours:
                values[i] = input_volume
        elif frequency == 'yearly':
            if day_of_year in active_days and hour_of_day in hours:
                values[i] = input_volume

    df = pd.DataFrame(values, index=period_index, columns=['value'], dtype=f"pint[{str(pint_unit)}]")

    return SourceHourlyValues(df, label="Hourly usage")


def create_hourly_usage_from_daily_volume_and_list_of_hours(
        timespan: pint.Quantity, daily_volume: float, hours: List[int],
        start_date: datetime = datetime.strptime("2025-01-01", "%Y-%m-%d"),
        pint_unit: pint.Unit = u.dimensionless):
    volume_per_hour = daily_volume / len(hours)

    return create_hourly_usage_from_frequency(
        timespan, volume_per_hour, frequency='daily', active_days=None, hours=hours,
        start_date=start_date, pint_unit=pint_unit)



from typing import Dict, List

import pandas as pd

from efootprint.abstract_modeling_classes.explainable_object_dict import ExplainableObjectDict
from efootprint.abstract_modeling_classes.list_linked_to_modeling_obj import ListLinkedToModelingObj
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.constants.units import u
from efootprint.core.hardware.network import Network
from efootprint.core.hardware.server import Server
from efootprint.core.hardware.storage import Storage
from efootprint.core.usage.usage_pattern import UsagePattern
from efootprint.core.usage.usage_journey import UsageJourney
from efootprint.abstract_modeling_classes.explainable_objects import ExplainableQuantity, EmptyExplainableObject, \
    ExplainableHourlyQuantities
from efootprint.logger import logger
from efootprint.utils.plot_emission_diffs import EmissionPlotter
from efootprint.utils.tools import format_co2_amount, display_co2_amount


class System(ModelingObject):
    def __init__(self, name: str, usage_patterns: List[UsagePattern]):
        super().__init__(name)
        self.total_footprint = EmptyExplainableObject()
        self.check_no_object_to_link_is_already_linked_to_another_system(usage_patterns)
        self.usage_patterns = ListLinkedToModelingObj(usage_patterns)
        self.previous_change = None
        self.previous_total_energy_footprints_sum_over_period = ExplainableObjectDict()
        self.previous_total_fabrication_footprints_sum_over_period = ExplainableObjectDict()
        self.all_changes = []
        self.initial_total_energy_footprints_sum_over_period = ExplainableObjectDict()
        self.initial_total_fabrication_footprints_sum_over_period = ExplainableObjectDict()
        self.simulation = None

    def compute_calculated_attributes(self):
        self.check_no_object_to_link_is_already_linked_to_another_system(self.usage_patterns)
        super().compute_calculated_attributes()

    @property
    def calculated_attributes(self) -> List[str]:
        return ["total_footprint"]

    @property
    def attributes_that_shouldnt_trigger_update_logic(self):
        return (super().attributes_that_shouldnt_trigger_update_logic
                + ["all_changes", "previous_change", "previous_total_energy_footprints_sum_over_period",
                   "previous_total_fabrication_footprints_sum_over_period",
                   "initial_total_energy_footprints_sum_over_period",
                   "initial_total_fabrication_footprints_sum_over_period", "simulation"])

    def check_no_object_to_link_is_already_linked_to_another_system(self, usage_patterns: List[UsagePattern]):
        for mod_obj in self.get_objects_linked_to_usage_patterns(usage_patterns):
            mod_obj_systems = mod_obj.systems
            if mod_obj_systems and mod_obj_systems[0].id != self.id:
                raise PermissionError(f"{mod_obj.name} is already linked to {mod_obj_systems[0].name}, so it is "
                                      f"impossible to link it to {self.name}")
            if len(mod_obj_systems) > 1:
                raise ValueError(f"{mod_obj.name} is linked to 2 systems, this should never happen, please report an"
                                 f" e-footprint bug at https://github.com/Boavizta/e-footprint/issues")

    @property
    def usage_journeys(self) -> List[UsageJourney]:
        output_set = set()
        for usage_pattern in self.usage_patterns:
            output_set.update({usage_pattern.usage_journey})

        return list(output_set)

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self):
        return self.usage_patterns

    @property
    def systems(self) -> List:
        return [self]

    def after_init(self):
        from time import time
        start = time()
        logger.info(f"Starting computing {self.name} modeling")
        mod_obj_computation_chain_excluding_self = self.mod_objs_computation_chain[1:]
        self.launch_mod_objs_computation_chain(mod_obj_computation_chain_excluding_self)
        self.compute_calculated_attributes()
        logger.info(f"Finished computing {self.name} modeling in {round(time() - start, 3)} seconds")
        self.initial_total_energy_footprints_sum_over_period = self.total_energy_footprint_sum_over_period
        self.initial_total_fabrication_footprints_sum_over_period = self.total_fabrication_footprint_sum_over_period
        self.trigger_modeling_updates = True

    def get_objects_linked_to_usage_patterns(self, usage_patterns: List[UsagePattern]):
        output_list =  self.storages_from_usage_patterns(usage_patterns) + usage_patterns + \
                      self.networks_from_usage_patterns(usage_patterns)
        usage_journeys = list(set([up.usage_journey for up in usage_patterns]))
        uj_steps = list(set(sum([uj.uj_steps for uj in usage_journeys], start=[])))
        jobs = list(set(sum([uj_step.jobs for uj_step in uj_steps], start=[])))
        devices = list(set(sum([up.devices for up in usage_patterns], start=[])))
        countries = list(set([up.country for up in usage_patterns]))
        servers = self.servers_from_usage_patterns(usage_patterns)
        services = list(set(sum([server.installed_services for server in servers], start=[])))
        all_modeling_objects = output_list + usage_journeys + uj_steps + jobs + devices + countries + servers + services

        return all_modeling_objects

    @property
    def all_linked_objects(self):
        return self.get_objects_linked_to_usage_patterns(self.usage_patterns)

    @staticmethod
    def servers_from_usage_patterns(usage_patterns: List[UsagePattern]) -> List[Server]:
        output_set = set()
        for usage_pattern in usage_patterns:
            output_set.update(usage_pattern.usage_journey.servers)

        return list(output_set)

    @property
    def servers(self) -> List[Server]:
        return self.servers_from_usage_patterns(self.usage_patterns)

    @staticmethod
    def storages_from_usage_patterns(usage_patterns: List[UsagePattern]) -> List[Storage]:
        output_set = set()
        for usage_pattern in usage_patterns:
            output_set.update(usage_pattern.usage_journey.storages)

        return list(output_set)

    @property
    def storages(self) -> List[Storage]:
        return self.storages_from_usage_patterns(self.usage_patterns)

    @staticmethod
    def networks_from_usage_patterns(usage_patterns: List[UsagePattern]) -> List[Network]:
        output_set = set()
        for usage_pattern in usage_patterns:
            output_set.update({usage_pattern.network})

        return list(output_set)

    @property
    def networks(self) -> List[Network]:
        return self.networks_from_usage_patterns(self.usage_patterns)

    def get_storage_by_name(self, storage_name) -> Storage:
        for storage in self.storages:
            if storage.name == storage_name:
                return storage

    def get_server_by_name(self, server_name) -> Server:
        for server in self.servers:
            if server.name == server_name:
                return server

    def get_usage_pattern_by_name(self, usage_pattern_name) -> UsagePattern:
        for usage_pattern in self.usage_patterns:
            if usage_pattern.name == usage_pattern_name:
                return usage_pattern

    @property
    def fabrication_footprints(self) -> Dict[str, Dict[str, ExplainableHourlyQuantities]]:
        fab_footprints = {
            "Servers": {server.name: server.instances_fabrication_footprint for server in self.servers},
            "Storage": {storage.name: storage.instances_fabrication_footprint for storage in self.storages},
            "Network": {"networks": EmptyExplainableObject()},
            "Devices": {usage_pattern.name: usage_pattern.instances_fabrication_footprint
                        for usage_pattern in self.usage_patterns},
        }

        return fab_footprints

    @property
    def energy_footprints(self) -> Dict[str, Dict[str, ExplainableHourlyQuantities]]:
        energy_footprints = {
            "Servers": {server.name: server.energy_footprint for server in self.servers},
            "Storage": {storage.name: storage.energy_footprint for storage in self.storages},
            "Network": {network.name: network.energy_footprint for network in self.networks},
            "Devices": {usage_pattern.name: usage_pattern.energy_footprint
                        for usage_pattern in self.usage_patterns},
        }

        return energy_footprints

    @property
    def total_fabrication_footprints(self) -> Dict[str, ExplainableHourlyQuantities]:
        fab_footprints = {
            "Servers": sum([server.instances_fabrication_footprint for server in self.servers],
                           start=EmptyExplainableObject()).to(u.kg).set_label(
                "Servers total fabrication footprint"),
            "Storage": sum([storage.instances_fabrication_footprint for storage in self.storages],
                           start=EmptyExplainableObject()).to(u.kg).set_label(
                "Storage total fabrication footprint"),
            "Network": EmptyExplainableObject(),
            "Devices": sum([usage_pattern.instances_fabrication_footprint
                           for usage_pattern in self.usage_patterns], start=EmptyExplainableObject()).to(u.kg).set_label(
                "Devices total fabrication footprint")
        }

        return fab_footprints

    @property
    def total_energy_footprints(self) -> Dict[str, ExplainableHourlyQuantities]:
        energy_footprints = {
            "Servers": sum([server.energy_footprint for server in self.servers], start=EmptyExplainableObject()
                           ).to(u.kg).set_label("Servers total energy footprint"),
            "Storage": sum([storage.energy_footprint for storage in self.storages], start=EmptyExplainableObject()
                           ).to(u.kg).set_label("Storage total energy footprint"),
            "Network": sum([network.energy_footprint for network in self.networks], start=EmptyExplainableObject()
                           ).to(u.kg).set_label("Network total energy footprint"),
            "Devices": sum([usage_pattern.energy_footprint for usage_pattern in self.usage_patterns],
                           start=EmptyExplainableObject()).to(u.kg).set_label("Devices total energy footprint")
        }

        return energy_footprints

    @staticmethod
    def sum_and_remove_empty_explainable_object(expl_obj):
        tmp_sum = expl_obj.sum()
        if isinstance(tmp_sum, EmptyExplainableObject):
            tmp_sum = ExplainableQuantity(0 * u.kg, "null value")

        return tmp_sum

    @property
    def fabrication_footprint_sum_over_period(self) -> Dict[str, Dict[ModelingObject, ExplainableQuantity]]:
        fab_footprints_sum = {}
        for category_key, category_dict in self.fabrication_footprints.items():
            fab_footprints_sum[category_key] = {
                obj_key: self.sum_and_remove_empty_explainable_object(obj_value).to(u.kg).set_label(
                    f"{obj_key} fabrication footprints summed over modeling period")
                for obj_key, obj_value in category_dict.items()
            }

        return fab_footprints_sum

    @property
    def energy_footprint_sum_over_period(self) -> Dict[str, Dict[ModelingObject, ExplainableQuantity]]:
        energy_footprints_sum = {}
        for key, dict_value in self.energy_footprints.items():
            energy_footprints_sum[key] = {
                obj_key: self.sum_and_remove_empty_explainable_object(obj_value).to(u.kg).set_label(
                    f"{obj_key} energy footprints summed over modeling period")
                for obj_key, obj_value in dict_value.items()
            }

        return energy_footprints_sum

    @property
    def total_fabrication_footprint_sum_over_period(self) -> Dict[str, ExplainableQuantity]:
        fab_footprints = {
            object_category: self.sum_and_remove_empty_explainable_object(category_value).to(u.kg).set_label(
                f"{object_category} total fabrication footprints summed over modeling period")
            for object_category, category_value in self.total_fabrication_footprints.items()
        }

        return ExplainableObjectDict(fab_footprints)

    @property
    def total_energy_footprint_sum_over_period(self) -> Dict[str, ExplainableQuantity]:
        energy_footprints = {
            object_category: self.sum_and_remove_empty_explainable_object(category_value).to(u.kg).set_label(
                f"{object_category} total energy footprints summed over modeling period")
            for object_category, category_value in self.total_energy_footprints.items()
        }

        return ExplainableObjectDict(energy_footprints)

    def update_total_footprint(self):
        total_footprint = (
            sum(
                sum(self.fabrication_footprints[key].values()) + sum(self.energy_footprints[key].values())
                for key in self.fabrication_footprints.keys()
            )
        ).to(u.kg).set_label(f"{self.name} total carbon footprint")

        self.total_footprint = round(total_footprint, 4)

    def plot_footprints_by_category_and_object(self, filename=None, height=400, width=800, return_only_html=False):
        import plotly.express as px
        import plotly

        fab_footprints = self.fabrication_footprint_sum_over_period
        energy_footprints = self.energy_footprint_sum_over_period
        categories = list(fab_footprints.keys())

        rows_as_dicts = []

        value_colname = "tonnes CO2 emissions"
        for category in categories:
            fab_objects = sorted(fab_footprints[category].items(), key=lambda x: x[0])
            energy_objects = sorted(energy_footprints[category].items(), key=lambda x: x[0])

            for objs, color in zip([energy_objects, fab_objects], ["Electricity", "Fabrication"]):
                data_dicts = [
                    {"Type": color, "Category": category, "Object": obj[0],
                     value_colname: obj[1].value.magnitude / 1000,
                     "Amount": f"{display_co2_amount(format_co2_amount(obj[1].value.magnitude))}"}
                    for obj in objs]
                rows_as_dicts += data_dicts

        df = pd.DataFrame.from_records(rows_as_dicts)

        total_co2 = df[value_colname].sum()

        total_footprint = self.total_footprint

        fig = px.bar(
            df, x="Category", y=value_colname, color='Type', barmode='group', height=height, width=width,
            hover_data={"Type": False, "Category": False, "Object": True, value_colname: False, "Amount": True},
            template="plotly_white",
            title=f"Total CO2 emissions from "
                  f"{total_footprint.value.index.min().to_timestamp().date()} "
                  f"to {total_footprint.value.index.max().to_timestamp().date()}"
                  f": {display_co2_amount(format_co2_amount(total_co2 * 1000, rounding_value=0))}"
                  )

        if (max(sum(energy_footprints["Servers"].values()), sum(fab_footprints["Servers"].values())) >
                max(sum(energy_footprints["Devices"].values()), sum(fab_footprints["Devices"].values()))):
            legend_alignment = "right"
            legend_x = 0.98
        else:
            legend_alignment = "left"
            legend_x = 0.02

        fig.update_layout(
            legend={"orientation": "v", "yanchor": "top", "y": 1.02, "xanchor": legend_alignment, "x": legend_x,
                    "title": ""},
            title={"x": 0.5, "y": 0.9, "xanchor": 'center', "yanchor": 'top'})

        total_co2_per_category_and_type = df.groupby(["Category", "Type"])[value_colname].sum()

        for category, source_type in total_co2_per_category_and_type.keys():
            height = total_co2_per_category_and_type.loc[category, source_type]
            x_shift_direction = 1 if source_type == 'Fabrication' else -1

            fig.add_annotation(
                x=category,
                y=height,
                text=f"{int((height / total_co2) * 100)}%",  # Format the label as a percentage
                showarrow=False,
                yshift=10,  # Shift the label slightly above the stack
                xshift=30 * x_shift_direction
            )

        if return_only_html:
            return fig.to_html(full_html=False, include_plotlyjs=False)

        else:
            from IPython.display import HTML

            if filename is None:
                filename = f"{self.name} footprints.html"

            plotly.offline.plot(fig, filename=filename, auto_open=False)

            return HTML(filename)

    def plot_emission_diffs(self, filepath=None, figsize=(10, 5), from_start=False, plt_show=False):
        from matplotlib import pyplot as plt

        if self.previous_change is None:
            raise ValueError(
                f"There has been no change to the system yet so no diff to plot.\n"
                f"Use System.plot_footprints_by_category_and_object() to visualize footprints")

        if from_start and len(self.all_changes) > 1:
            emissions_dict__old = [self.initial_total_energy_footprints_sum_over_period,
                                   self.initial_total_fabrication_footprints_sum_over_period]
        else:
            emissions_dict__old = [self.previous_total_energy_footprints_sum_over_period,
                                   self.previous_total_fabrication_footprints_sum_over_period]

        emissions_dict__new = [self.total_energy_footprint_sum_over_period,
                               self.total_fabrication_footprint_sum_over_period]

        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=figsize)

        EmissionPlotter(
            ax, emissions_dict__old, emissions_dict__new, rounding_value=0).plot_emission_diffs()

        if filepath is not None:
            plt.savefig(filepath, bbox_inches='tight')

        if plt_show:
            plt.show()



from typing import Type


class ObjectLinkedToModelingObj:
    def __init__(self):
        self.modeling_obj_container = None
        self.attr_name_in_mod_obj_container = None

    def set_modeling_obj_container(
            self, new_parent_modeling_object: Type["ModelingObject"] | None, attr_name: str | None):
        if new_parent_modeling_object is None or attr_name is None:
            assert new_parent_modeling_object == attr_name, (
                f"Both new_parent_modeling_object and attr_name should be None or not None. "
                f"Here new_parent_modeling_object is {new_parent_modeling_object} and attr_name is {attr_name}.")
        if (self.modeling_obj_container is not None and new_parent_modeling_object is not None and
                new_parent_modeling_object.id != self.modeling_obj_container.id):
            raise ValueError(
                f"A {self.__class__.__name__} can’t be attributed to more than one ModelingObject. Here "
                f"{self} is trying to be linked to {new_parent_modeling_object.name} but is already linked to "
                f"{self.modeling_obj_container.name}.")
        self.modeling_obj_container = new_parent_modeling_object
        self.attr_name_in_mod_obj_container = attr_name

    @property
    def id(self):
        if self.modeling_obj_container is None:
            raise ValueError(
                f"{self} doesn’t have a modeling_obj_container, hence it makes no sense "
                f"to look for its ancestors")

        return f"{self.attr_name_in_mod_obj_container}-in-{self.modeling_obj_container.id}"

    @property
    def dict_container(self):
        output = None
        if (
                self.modeling_obj_container is not None
                and isinstance(getattr(self.modeling_obj_container, self.attr_name_in_mod_obj_container), dict)
                and id(getattr(self.modeling_obj_container, self.attr_name_in_mod_obj_container)) != id(self)
        ):
            output = getattr(self.modeling_obj_container, self.attr_name_in_mod_obj_container)

        return output

    @property
    def key_in_dict(self):
        if self.dict_container is None:
            raise ValueError(f"{self} is not linked to a ModelingObject through a dictionary attribute.")
        else:
            output_key = None
            for key, value in self.dict_container.items():
                if id(value) == id(self):
                    if output_key is None:
                        output_key = key
                    else:
                        raise ValueError(f"Multiple keys found for {self} in {self.dict_container}.")

        return output_key

    @property
    def list_container(self):
        output = None
        if (
                not isinstance(self, list)
                and self.modeling_obj_container is not None
                and isinstance(getattr(self.modeling_obj_container, self.attr_name_in_mod_obj_container), list)
        ):
            output = getattr(self.modeling_obj_container, self.attr_name_in_mod_obj_container)

        return output

    @property
    def indexes_in_list(self):
        if self.list_container is None:
            raise ValueError(f"{self} is not linked to a ModelingObject through a list attribute.")
        else:
            output_indexes = []
            for index, value in enumerate(self.list_container):
                if id(value) == id(self):
                    output_indexes.append(index)

        return output_indexes

    def replace_in_mod_obj_container_without_recomputation(self, new_value):
        assert self.modeling_obj_container is not None, f"{self} is not linked to a ModelingObject."
        assert isinstance(new_value, ObjectLinkedToModelingObj), (
            f"Trying to replace {self} by {new_value} which is not an instance of "
            f"ObjectLinkedToModelingObj.")
        from efootprint.abstract_modeling_classes.explainable_objects import EmptyExplainableObject

        if not isinstance(new_value, EmptyExplainableObject) and not isinstance(self, EmptyExplainableObject):
            assert isinstance(new_value, self.__class__) or isinstance(self, new_value.__class__), \
                f"Trying to replace {self} by {new_value} which is of different type."
        mod_obj_container = self.modeling_obj_container
        attr_name = self.attr_name_in_mod_obj_container
        if self.dict_container is not None:
            if self.key_in_dict not in self.dict_container.keys():
                raise KeyError(f"object of id {self.key_in_dict.id} not found as key in {attr_name} attribute of "
                               f"{mod_obj_container.id} when trying to replace {self} by {new_value}. "
                               f"This should not happen.")
            self.dict_container[self.key_in_dict] = new_value
        elif self.list_container is not None:
            if not self.indexes_in_list:
                raise ValueError(f"object of id {self.id} not found in {attr_name} attribute of {mod_obj_container.id} "
                                 f"when trying to replace \n\n{self}\nby\n\n{new_value}.\n\nThis should not happen.")
            for index in self.indexes_in_list:
                initial_trigger_modeling_updates = self.list_container.trigger_modeling_updates
                self.list_container.trigger_modeling_updates = False
                self.list_container[index] = new_value
                self.list_container.trigger_modeling_updates = initial_trigger_modeling_updates
        else:
            mod_obj_container.__dict__[attr_name] = new_value
        self.set_modeling_obj_container(None, None)
        new_value.set_modeling_obj_container(mod_obj_container, attr_name)



import math
from typing import List, Type

import numpy as np
import pandas as pd
import pint_pandas

from efootprint.builders.time_builders import create_hourly_usage_df_from_list
from efootprint.constants.sources import Sources
from efootprint.core.hardware.infra_hardware import InfraHardware
from efootprint.abstract_modeling_classes.explainable_objects import ExplainableQuantity, ExplainableHourlyQuantities, \
    EmptyExplainableObject
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u


class Storage(InfraHardware):
    @classmethod
    def default_values(cls):
        return {
            "carbon_footprint_fabrication_per_storage_capacity": SourceValue(160 * u.kg / u.TB),
            "power_per_storage_capacity": SourceValue(1.3 * u.W / u.TB),
            "lifespan": SourceValue(6 * u.years),
            "idle_power": SourceValue(0 * u.W),
            "storage_capacity": SourceValue(1 * u.TB),
            "data_replication_factor": SourceValue(3 * u.dimensionless),
            "base_storage_need": SourceValue(0 * u.TB),
            "data_storage_duration": SourceValue(5 * u.year)
        }

    @classmethod
    def ssd(cls, name="Default SSD storage", **kwargs):
        output_args = {
            "carbon_footprint_fabrication_per_storage_capacity": SourceValue(
                160 * u.kg / u.TB, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            "power_per_storage_capacity": SourceValue(1.3 * u.W / u.TB, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            "lifespan": SourceValue(6 * u.years, Sources.HYPOTHESIS),
            "idle_power": SourceValue(0 * u.W, Sources.HYPOTHESIS),
            "storage_capacity": SourceValue(1 * u.TB, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            "data_replication_factor": SourceValue(3 * u.dimensionless, Sources.HYPOTHESIS),
            "base_storage_need": SourceValue(0 * u.TB, Sources.HYPOTHESIS),
            "data_storage_duration": SourceValue(5 * u.year, Sources.HYPOTHESIS)
        }

        output_args.update(kwargs)

        return cls(name, **output_args)

    @classmethod
    def hdd(cls, name="Default HDD storage", **kwargs):
        output_args = {
            "carbon_footprint_fabrication_per_storage_capacity": SourceValue(
                20 * u.kg / u.TB, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            "power_per_storage_capacity": SourceValue(4.2 * u.W / u.TB, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            "lifespan": SourceValue(4 * u.years, Sources.HYPOTHESIS),
            "idle_power": SourceValue(0 * u.W, Sources.HYPOTHESIS),
            "storage_capacity": SourceValue(1 * u.TB, Sources.STORAGE_EMBODIED_CARBON_STUDY),
            "data_replication_factor": SourceValue(3 * u.dimensionless, Sources.HYPOTHESIS),
            "base_storage_need": SourceValue(0 * u.TB, Sources.HYPOTHESIS),
            "data_storage_duration": SourceValue(5 * u.year, Sources.HYPOTHESIS)
        }

        output_args.update(kwargs)

        return cls(name, **output_args)

    @classmethod
    def archetypes(cls):
        return [cls.ssd, cls.hdd]

    def __init__(self, name: str, carbon_footprint_fabrication_per_storage_capacity: ExplainableQuantity,
                 power_per_storage_capacity: ExplainableQuantity, lifespan: ExplainableQuantity,
                 idle_power: ExplainableQuantity,
                 storage_capacity: ExplainableQuantity, data_replication_factor: ExplainableQuantity,
                 data_storage_duration: ExplainableQuantity, base_storage_need: ExplainableQuantity,
                 fixed_nb_of_instances: ExplainableQuantity | EmptyExplainableObject = None):
        super().__init__(
            name, carbon_footprint_fabrication=SourceValue(0 * u.kg), power=SourceValue(0 * u.W), lifespan=lifespan)
        self.carbon_footprint_fabrication_per_storage_capacity = (carbon_footprint_fabrication_per_storage_capacity
        .set_label(f"Fabrication carbon footprint of {self.name} per storage capacity"))
        self.power_per_storage_capacity = power_per_storage_capacity.set_label(
            f"Power of {self.name} per storage capacity")
        self.idle_power = idle_power.set_label(f"Idle power of {self.name}")
        self.storage_capacity = storage_capacity.set_label(f"Storage capacity of {self.name}")
        self.data_replication_factor = data_replication_factor.set_label(f"Data replication factor of {self.name}")
        self.data_storage_duration = data_storage_duration.set_label(f"Data storage duration of {self.name}")
        self.base_storage_need = base_storage_need.set_label(f"{self.name} initial storage need")
        self.fixed_nb_of_instances = (fixed_nb_of_instances or EmptyExplainableObject()).set_label(
            f"User defined number of {self.name} instances").to(u.dimensionless)
        self.storage_delta = EmptyExplainableObject()
        self.full_cumulative_storage_need = EmptyExplainableObject()
        self.nb_of_active_instances = EmptyExplainableObject()

    @property
    def server(self) -> Type["Server"]:
        if self.modeling_obj_containers:
            if len(self.modeling_obj_containers) > 1:
                raise PermissionError(
                    f"Storage object can only be associated with one server object but {self.name} is associated "
                    f"with {[mod_obj.name for mod_obj in self.modeling_obj_containers]}")
            return self.modeling_obj_containers[0]
        else:
            return None

    @property
    def calculated_attributes(self):
        return (
            ["carbon_footprint_fabrication", "power", "storage_delta", "full_cumulative_storage_need",
             "raw_nb_of_instances", "nb_of_instances", "nb_of_active_instances", "instances_fabrication_footprint",
             "instances_energy", "energy_footprint"])

    @property
    def jobs(self) -> List[Type["Job"]]:
        return list(set(
            job for serv in self.modeling_obj_containers for job in serv.jobs
        ))

    @property
    def power_usage_effectiveness(self):
        if self.server is not None:
            return self.server.power_usage_effectiveness
        else:
            return EmptyExplainableObject()

    @property
    def average_carbon_intensity(self):
        if self.server is not None:
            return self.server.average_carbon_intensity
        else:
            return EmptyExplainableObject()

    def update_carbon_footprint_fabrication(self):
        self.carbon_footprint_fabrication = (
                self.carbon_footprint_fabrication_per_storage_capacity * self.storage_capacity).set_label(
            f"Carbon footprint of {self.name}")

    def update_power(self):
        self.power = (self.power_per_storage_capacity * self.storage_capacity).set_label(f"Power of {self.name}")

    # If storage_needed, storage_freed and automatic_storage_dumps_after_storage_duration had their update function
    # and were attributes of the storage class then the update of the data_stored attribute of a job from positive to
    # negative or negative to positive would cause problem with the re-computation of modeling attributes.
    # For example if the data_stored attribute of a job went from negative to positive then the storage_freed would be
    # recomputed but not the storage_needed nor the automatic_storage_dumps_after_storage_duration.
    # By turning these three attributes into properties, we make all of them dependencies of the calculation of
    # storage_delta and solve the problem.
    @property
    def storage_needed(self):
        storage_needed = EmptyExplainableObject()

        for job in self.jobs:
            if job.data_stored.magnitude >= 0:
                storage_needed += job.hourly_data_stored_across_usage_patterns

        storage_needed *= self.data_replication_factor

        return storage_needed.to(u.TB).set_label(f"Hourly {self.name} storage need")

    @property
    def storage_freed(self):
        storage_freed = EmptyExplainableObject()

        for job in self.jobs:
            if job.data_stored.magnitude < 0:
                storage_freed += job.hourly_data_stored_across_usage_patterns

        storage_freed *= self.data_replication_factor

        return storage_freed.to(u.TB).set_label(f"Hourly {self.name} storage freed")

    @property
    def automatic_storage_dumps_after_storage_duration(self):
        if isinstance(self.storage_needed, EmptyExplainableObject):
            return EmptyExplainableObject(left_parent=self.storage_needed)
        else:
            storage_duration_in_hours = math.ceil(self.data_storage_duration.to(u.hour).magnitude)
            automatic_storage_dumps_after_storage_duration_df = - self.storage_needed.value.copy().shift(
                periods=storage_duration_in_hours, freq='h')
            automatic_storage_dumps_after_storage_duration_df = automatic_storage_dumps_after_storage_duration_df[
                automatic_storage_dumps_after_storage_duration_df.index <= self.storage_needed.value.index.max()]

            if len(automatic_storage_dumps_after_storage_duration_df) == 0:
                storage_needs_start_date = self.storage_needed.value.index.min().to_timestamp()
                storage_needs_end_date = self.storage_needed.value.index.max().to_timestamp()
                storage_needs_nb_of_hours = int((storage_needs_end_date - storage_needs_start_date).seconds / 3600)
                automatic_storage_dumps_after_storage_duration_df = create_hourly_usage_df_from_list(
                    [0] * (storage_needs_nb_of_hours + 1), start_date=storage_needs_start_date)

            return ExplainableHourlyQuantities(
                automatic_storage_dumps_after_storage_duration_df, label=f"Storage dumps for {self.name}",
                left_parent=self.storage_needed,
                right_parent=self.data_storage_duration, operator="shift by storage duration and negate")

    def update_storage_delta(self):
        storage_delta = (self.storage_needed + self.storage_freed
                         + self.automatic_storage_dumps_after_storage_duration)

        self.storage_delta = storage_delta.set_label(f"Hourly storage delta for {self.name}")

    def update_full_cumulative_storage_need(self):
        if isinstance(self.storage_delta, EmptyExplainableObject):
            self.full_cumulative_storage_need = EmptyExplainableObject(left_parent=self.storage_delta)
        else:
            storage_delta_df = self.storage_delta.value.copy()
            storage_delta_df.iat[0, 0] += self.base_storage_need.value
            full_cumulative_storage_need = storage_delta_df.cumsum()

            if full_cumulative_storage_need.value.min().magnitude < 0:
                jobs_in_errors = [
                    f"name: {job.name} - value: {job.data_stored}"
                    for job in self.jobs if job.data_stored.magnitude < 0]
                raise ValueError(
                    f"In Storage object {self.name}, negative cumulative storage need detected: "
                    f"{full_cumulative_storage_need.min().value}."
                    f"Please verify your jobs that delete data: {jobs_in_errors}"
                    f" or increase the base_storage_need value, currently set to {self.base_storage_need.value}"
                )

            self.full_cumulative_storage_need = ExplainableHourlyQuantities(
                full_cumulative_storage_need, label=f"Full cumulative storage need for {self.name}",
                left_parent=self.storage_delta, right_parent=self.base_storage_need,
                operator="cumulative sum of storage delta with initial storage need")

    def update_raw_nb_of_instances(self):
        raw_nb_of_instances = (self.full_cumulative_storage_need / self.storage_capacity).to(u.dimensionless)

        self.raw_nb_of_instances = raw_nb_of_instances.set_label(f"Hourly raw number of instances for {self.name}")

    def update_nb_of_instances(self):
        if isinstance(self.raw_nb_of_instances, EmptyExplainableObject):
            self.nb_of_instances = EmptyExplainableObject(left_parent=self.raw_nb_of_instances)
        else:
            nb_of_instances = self.raw_nb_of_instances.ceil()

            if not isinstance(self.fixed_nb_of_instances, EmptyExplainableObject):
                max_nb_of_instances = nb_of_instances.max()
                if max_nb_of_instances > self.fixed_nb_of_instances:
                    raise ValueError(
                        f"The number of {self.name} instances computed from its resources need is superior to the "
                        f"number of instances specified by the user/server "
                        f"({max_nb_of_instances} > {self.fixed_nb_of_instances})")
                else:
                    fixed_nb_of_instances_df = pd.DataFrame(
                        {"value": pint_pandas.PintArray(
                            np.full(len(self.raw_nb_of_instances),
                                    self.fixed_nb_of_instances.to(u.dimensionless).magnitude),
                            dtype=u.dimensionless
                        )},
                        index=self.raw_nb_of_instances.value.index
                    )
                    fixed_nb_of_instances = ExplainableHourlyQuantities(
                        fixed_nb_of_instances_df,
                        "Nb of instances",
                        left_parent=self.raw_nb_of_instances,
                        right_parent=self.fixed_nb_of_instances
                    )
                self.nb_of_instances = fixed_nb_of_instances.set_label(
                    f"Hourly fixed number of instances for {self.name}")
            else:
                nb_of_instances = ExplainableHourlyQuantities(
                    nb_of_instances.value, left_parent=nb_of_instances,
                    right_parent=self.fixed_nb_of_instances, operator="depending on being empty")
                self.nb_of_instances = nb_of_instances.set_label(f"Hourly number of instances for {self.name}")

    def update_nb_of_active_instances(self):
        tmp_nb_of_active_instances = (
                (self.storage_needed.abs().np_compared_with(self.storage_freed.abs(), "max")
                 + self.automatic_storage_dumps_after_storage_duration.abs())
                / self.storage_capacity
        ).to(u.dimensionless)
        nb_of_active_instances = tmp_nb_of_active_instances.np_compared_with(self.nb_of_instances.abs(), "min")

        self.nb_of_active_instances = nb_of_active_instances.set_label(
            f"Hourly number of active instances for {self.name}")

    def update_instances_energy(self):
        nb_of_idle_instances = (self.nb_of_instances - self.nb_of_active_instances).set_label(
            f"Hourly number of idle instances for {self.name}")
        active_storage_energy = (
                self.nb_of_active_instances * self.power * ExplainableQuantity(
            1 * u.hour, "one hour") * self.power_usage_effectiveness
        ).set_label(f"Hourly active instances energy for {self.name}")
        idle_storage_energy = (
                nb_of_idle_instances * self.idle_power * ExplainableQuantity(
            1 * u.hour, "one hour") * self.power_usage_effectiveness
        ).set_label(f"Hourly idle instances energy for {self.name}")

        storage_energy = (active_storage_energy + idle_storage_energy)

        self.instances_energy = storage_energy.to(u.kWh).set_label(f"Storage energy for {self.name}")



from pyvis.network import Network

from efootprint.utils.graph_tools import WIDTH, HEIGHT, set_string_max_width

COLOR_MAP = {
    "Server": "red",
    "BoaviztaCloudServer": "red",
    "GPUServer": "red",
    "Device": "red",
    "Storage": "red",
    "UsagePattern": "blue",
    "UsageJourney": "dodgerblue",
    "UsageJourneyStep": "deepskyblue",
    "Job": "palegoldenrod",
    "VideoStreamingJob": "palegoldenrod",
    "GenAIJob": "palegoldenrod",
    "WebApplicationJob": "palegoldenrod",
    "VideoStreaming": "orange",
    "GenAIModel": "orange",
    "WebApplication": "orange",
}

USAGE_PATTERN_VIEW_CLASSES_TO_IGNORE = [
    "System", "Network", "Device", "Country", "Job", "Storage", "VideoStreaming", "GenAIModel", "WebApplication",
    "VideoStreamingJob", "GenAIJob", "WebApplicationJob"]
INFRA_VIEW_CLASSES_TO_IGNORE = [
    "UsagePattern", "Network", "Device", "System", "UsageJourney", "UsageJourneyStep", "Country"]


def build_object_relationships_graph(
        input_mod_obj, input_graph=None, visited_python_ids=None, classes_to_ignore=None, width=WIDTH, height=HEIGHT,
        notebook=False):
    cdn_resources = "local"
    if notebook:
        cdn_resources = "in_line"
    if classes_to_ignore is None:
        classes_to_ignore = ["System"]
    if input_graph is None:
        input_graph = Network(notebook=notebook, width=width, height=height, cdn_resources=cdn_resources)
    if visited_python_ids is None:
        visited_python_ids = set()

    if id(input_mod_obj) in visited_python_ids:
        return input_graph
    # Python ids are used to track visited objects because there can be multiple ContextualModelingObjectAttributes
    # pointing to the same ModelingObject instance, and their ids will be the same in this case, whereas the Python ids
    # will always be different.
    visited_python_ids.add(id(input_mod_obj))

    input_mod_obj_type = input_mod_obj.class_as_simple_str
    if input_mod_obj_type not in classes_to_ignore:
        input_graph.add_node(
            input_mod_obj.id, label=set_string_max_width(f"{input_mod_obj.name}", 20),
            title=set_string_max_width(str(input_mod_obj), 80),
            color=COLOR_MAP.get(input_mod_obj_type, "gray"))

    for mod_obj_attribute in input_mod_obj.mod_obj_attributes:
        mod_obj_attribute_type = mod_obj_attribute.class_as_simple_str
        if mod_obj_attribute_type not in classes_to_ignore:
            input_graph.add_node(
                mod_obj_attribute.id, label=set_string_max_width(f"{mod_obj_attribute.name}", 20),
                title=set_string_max_width(str(mod_obj_attribute), 80),
                color=COLOR_MAP.get(mod_obj_attribute_type, "gray"))
            if input_mod_obj_type not in classes_to_ignore:
                input_graph.add_edge(input_mod_obj.id, mod_obj_attribute.id)
            else:
                recursively_create_link_with_latest_non_ignored_node(
                    input_mod_obj, mod_obj_attribute, input_graph, classes_to_ignore)

        if mod_obj_attribute not in visited_python_ids:
            build_object_relationships_graph(mod_obj_attribute, input_graph, visited_python_ids, classes_to_ignore, width, height)

    return input_graph


def recursively_create_link_with_latest_non_ignored_node(source_obj, new_obj_to_link, input_graph, classes_to_ignore):
    for mod_obj in source_obj.modeling_obj_containers:
        if mod_obj.class_as_simple_str not in classes_to_ignore:
            if mod_obj.id != new_obj_to_link.id and mod_obj.id in input_graph.get_nodes():
                input_graph.add_edge(mod_obj.id, new_obj_to_link.id)
        else:
            recursively_create_link_with_latest_non_ignored_node(
                mod_obj, new_obj_to_link, input_graph, classes_to_ignore)



from typing import List, Type

from efootprint.abstract_modeling_classes.explainable_objects import ExplainableQuantity
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.abstract_modeling_classes.list_linked_to_modeling_obj import ListLinkedToModelingObj
from efootprint.constants.units import u
from efootprint.core.usage.job import JobBase


class UsageJourneyStep(ModelingObject):
    @classmethod
    def default_values(cls):
        return {"user_time_spent": SourceValue(1 * u.min)}

    def __init__(self, name: str, user_time_spent: ExplainableQuantity, jobs: List[JobBase]):
        super().__init__(name)
        self.user_time_spent = user_time_spent
        self.user_time_spent.set_label(f"Time spent on step {self.name}")
        self.jobs = ListLinkedToModelingObj(jobs)

    @property
    def usage_journeys(self) -> List[Type["UsageJourney"]]:
        return self.modeling_obj_containers

    @property
    def usage_patterns(self) -> List[Type["UsagePattern"]]:
        return list(set(sum([uj.usage_patterns for uj in self.usage_journeys], start=[])))

    @property
    def networks(self) -> List[Type["Network"]]:
        return list(set([up.network for up in self.usage_patterns]))

    @property
    def systems(self) -> List:
        return list(set(sum([up.systems for up in self.usage_patterns], start=[])))

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List[Type["UsageJourney"]]:
        return self.jobs + self.networks
